/*
 * jMax
 * Copyright (C) 1994, 1995, 1998, 1999 by IRCAM-Centre Georges Pompidou, Paris, France.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * See file LICENSE for further informations on licensing terms.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <fts/fts.h>
#include <utils/c/include/utils.h>
#include <data/c/include/ivec.h>

#include <stdlib.h>

static fts_symbol_t sym_text = 0;

fts_symbol_t ivec_symbol = 0;
fts_class_t *ivec_type = 0;

static fts_symbol_t sym_local = 0;

static fts_symbol_t sym_set_visibles = 0;
static fts_symbol_t sym_append_visibles = 0;
static fts_symbol_t sym_set_pixels = 0;
static fts_symbol_t sym_add_pixels = 0;
static fts_symbol_t sym_append_pixels = 0;
static fts_symbol_t sym_set_size = 0;
static fts_symbol_t sym_end_edit = 0;
static fts_symbol_t sym_start_edit = 0;
static fts_symbol_t sym_set_visible_size = 0;
static fts_symbol_t sym_copy_client = 0;
static fts_symbol_t sym_paste_client = 0;
static fts_symbol_t sym_cut_client = 0;
static fts_symbol_t sym_insert_client = 0;

/********************************************************
 *
 *  utility functions
 *
 */

#define ivec_set_editor_open(b) ((b)->opened = 1)
#define ivec_set_editor_close(b) ((b)->opened = 0)
#define ivec_editor_is_open(b) ((b)->opened)

void
ivec_set_size(ivec_t *vec, int size)
{
  int i;

  if(size > vec->alloc)
    {
      vec->values = (int *)fts_realloc(vec->values, sizeof(int) * size);
      vec->alloc = size;
    }

  /* when shortening: zero old values */
  for(i=size; i<vec->size; i++)
    vec->values[i] = 0;

  vec->size = size;
}

void
ivec_set_const(ivec_t *vec, int c)
{
  int *values = vec->values;
  int i;
  
  for(i=0; i<vec->size; i++)
    values[i] = c;
}

void
ivec_set_with_onset_from_atoms(ivec_t *vec, int onset, int ac, const fts_atom_t *at)
{
  int size = ivec_get_size(vec);
  int i;
 
  if(onset + ac > size)
    ac = size - onset;
  
  for(i=0; i<ac; i++)
    {
      if(fts_is_number(at + i))
	vec->values[i + onset] = fts_get_number_int(at + i);
      else
	vec->values[i + onset] = 0;
    }
}

/* sum, min, max */
int 
ivec_get_sum(ivec_t *vec)
{
  int sum = 0;
  int i;

  for(i=0; i<vec->size; i++)
    sum += vec->values[i];

  return sum;
}

int
ivec_get_sub_sum(ivec_t *vec, int from, int to)
{
  int sum = 0;
  int i;
  
  if(from < 0)
    from = 0;

  if(to >= vec->size)
    to = vec->size - 1;

  for(i=from; i<=to; i++)
    sum += vec->values[i];

  return sum;
}

int
ivec_get_min_value(ivec_t *vec)
{
  int min;
  int i;

  min = vec->values[0];

  for (i=1; i<vec->size; i++)
    if (vec->values[i] < min)
      min = vec->values[i];

  return min;
}


int
ivec_get_max_value(ivec_t *vec)
{
  int max;
  int i;

  max = vec->values[0];

  for (i=1; i<vec->size; i++)
    if (vec->values[i] > max)
      max = vec->values[i];

  return max;
}

void
ivec_copy(ivec_t *org, ivec_t *copy)
{
  int i;
  
  ivec_set_size(copy, org->size);

  for(i=0; i<org->size; i++)
    copy->values[i] = org->values[i];
}

/********************************************************
 *
 *  files
 *
 */

#define IVEC_BLOCK_SIZE 256

static void
ivec_grow(ivec_t *vec, int size)
{
  int alloc = vec->alloc;

  while(!alloc || size > alloc)
    alloc += IVEC_BLOCK_SIZE;

  ivec_set_size(vec, alloc);
}

int 
ivec_read_atom_file(ivec_t *vec, fts_symbol_t file_name)
{
  fts_atom_file_t *file = fts_atom_file_open(file_name, "r");
  int n = 0;
  fts_atom_t a;
  char c;

  if(!file)
    return -1;
  
  while(fts_atom_file_read(file, &a, &c))
    {
      if(n >= vec->alloc)
	ivec_grow(vec, n);

      if(fts_is_number(&a))
	ivec_set_element(vec, n, fts_get_number_int(&a));
      else
	ivec_set_element(vec, n, 0);
	
      n++;
    }

  ivec_set_size(vec, n);
  
  fts_atom_file_close(file);

  return (n);
}

int
ivec_write_atom_file(ivec_t *vec, fts_symbol_t file_name)
{
  fts_atom_file_t *file;
  int size = ivec_get_size(vec);
  int i;

  file = fts_atom_file_open(file_name, "w");

  if(!file)
    return -1;

  /* write the content of the vec */
  for(i=0; i<size; i++)     
    {
      fts_atom_t a;
      
      fts_set_int(&a, ivec_get_element(vec, i));
      fts_atom_file_write(file, &a, '\n');
    }

  fts_atom_file_close(file);

  return (i);
}

/********************************************************************
 *
 *   user methods
 *
 */
#define IVEC_CLIENT_BLOCK_SIZE 256

static void
ivec_send_visibles(ivec_t *ivec)
{
  int i;
  fts_atom_t a[IVEC_CLIENT_BLOCK_SIZE];
  int vecsize = ivec_get_size(ivec);
  int n = (ivec->vindex+ivec->vsize <= vecsize)? (ivec->vindex + ivec->vsize) : vecsize;

  int append = 0;
  int count = 0; 
  int send = 0;
  int current = 0;
  int veconset = 2;

  while(n > 0)
    {
      if(!append)
	{
	  fts_set_int(&a[0], vecsize);
	  fts_set_int(&a[1], n);
	}
      else
	{
	  fts_set_int(&a[0], count);
	  veconset = 1;
	}
      send = (n > IVEC_CLIENT_BLOCK_SIZE-veconset)? IVEC_CLIENT_BLOCK_SIZE-veconset: n;

      for(i = 0; ((i < send)&&(current+i<vecsize)); i++)
	  fts_set_int(&a[i+veconset], ivec->values[current+i]);

      if(!append)
	{
	  fts_client_send_message((fts_object_t *)ivec, sym_set_visibles, send+2, a);
	  append = 1;
	}
      else
	fts_client_send_message((fts_object_t *)ivec, sym_append_visibles, send+1, a);

      current+=send;
      count+=send;
      n -= send;
    }
}

static void
ivec_send_pixels(ivec_t *ivec)
{
  int i;
  fts_atom_t a[IVEC_CLIENT_BLOCK_SIZE];
  int vecsize = ivec_get_size(ivec);
  int n = ivec->pixsize;
  float k = (1/ivec->zoom);

  int append = 0;
  int count = 0; 
  int send = 0;
  /*int current = 0;*/
  int current = ivec->vindex;

  while(n > 0)
    {
      if(!append)
	  fts_set_int(&a[0], n);
      else
	  fts_set_int(&a[0], count);

      send = (n > IVEC_CLIENT_BLOCK_SIZE-1)? (IVEC_CLIENT_BLOCK_SIZE-1): n;

      for(i = 0; ((i < send)&&((int)(current+i*k)<vecsize)); i++)
	  fts_set_int(&a[i+1], ivec->values[(int)(current+k*i)]);
      
      if(!append)
	  {
	      fts_client_send_message((fts_object_t *)ivec, sym_set_pixels, send+1, a);
	      append = 1;
	  }
      else
	  fts_client_send_message((fts_object_t *)ivec, sym_append_pixels, send+1, a);

      current+=k*send;
      count+=send;
      n -= send;
    }
}

static void
ivec_insert_pixels(ivec_t *ivec, int startId, int size)
{
  int i;
  fts_atom_t a[IVEC_CLIENT_BLOCK_SIZE];
  int vecsize = ivec_get_size(ivec);
  float k = (1 / ivec->zoom);
  int send = (int)(size * ivec->zoom) + 1;
  int current = startId;

  fts_set_int(&a[0], (int)((startId - ivec->vindex) * ivec->zoom));
  
  for(i = 0; ((i < send) && ((int)(current + i * k) < vecsize)); i++)
    fts_set_int(&a[i+1], ivec->values[(int)(current+k*i)]);
      
  fts_client_send_message((fts_object_t *)ivec, sym_append_pixels, send + 1, a);
}

static void
ivec_append_visibles(ivec_t *ivec, int first, int last)
{
  int i;
  fts_atom_t a[IVEC_CLIENT_BLOCK_SIZE];

  int n = (last-first);

  int current = first;

  while(n > 0)
    {
	int send = (n > IVEC_CLIENT_BLOCK_SIZE-1)? IVEC_CLIENT_BLOCK_SIZE-1: n;

	fts_set_int(&a[0], current);
	
	for(i = 0; i < send; i++)
	    fts_set_int(&a[i+1], ivec->values[current+i]);
      
	fts_client_send_message((fts_object_t *)ivec, sym_append_visibles, send+1, a);

	current+=send;
	n -= send;
    }
}

static void
ivec_append_pixels(ivec_t *ivec, int deltax, int deltap)
{
  int i;
  fts_atom_t a[IVEC_CLIENT_BLOCK_SIZE];
  int vecsize = ivec_get_size(ivec);
  float k = (1/ivec->zoom);
  int n = (deltax > 0)? deltax : -deltax;

  int current = (deltax < 0)? ivec->vindex : (ivec->vindex+ivec->vsize-deltap);
  int start = (deltax < 0)? 0 : ivec->pixsize-deltax;

  while(n > 0)
    {
	int send = (n > IVEC_CLIENT_BLOCK_SIZE-1)? IVEC_CLIENT_BLOCK_SIZE-1: n;
	
	fts_set_int(&a[0], start);

	for(i = 0; ((i < send)&&((int)(current+i*k) < vecsize)); i++)
	    fts_set_int(&a[i+1], ivec->values[(int)(current+k*i)]);
      
	fts_client_send_message((fts_object_t *)ivec, sym_add_pixels, send+1, a);

	current+=k*i;
	start+=send;
	n -= send;
    }
}

static void
ivec_fill(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int constant = fts_get_int_arg(ac, at, 0, 0);

  ivec_set_const(this, constant);

  if(ivec_editor_is_open(this))
    {
      if(this->zoom < 0.5) 
	ivec_send_pixels(this);

      ivec_send_visibles(this);
    }
}

static void
ivec_set_elements(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  if(ac > 1 && fts_is_number(at))
    {
      int size = ivec_get_size(this);
      int onset = fts_get_number_int(at);

      if(onset >= 0 && onset < size)
	{
	  ivec_set_with_onset_from_atoms(this, onset, ac - 1, at + 1);
	  
	  if(ivec_editor_is_open(this))
	    {
	      if(this->zoom < 0.5) 
		ivec_insert_pixels(this, onset, ac - 1);
	      
	      fts_client_send_message((fts_object_t *)this, sym_append_visibles, ac, at);
	    }
	}
    }
}

static void
ivec_reverse(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int *ptr = ivec_get_ptr(this);
  int size = ivec_get_size(this); 
  int i, j;

  for(i=0, j=size-1; i<size/2; i++, j--)
    {
      int f = ptr[i];

      ptr[i] = ptr[j];
      ptr[j] = f;
    }
  
}

static int 
ivec_element_compare(const void *left, const void *right)
{
  int l = *((const int *)left);
  int r = *((const int *)right);

  return l - r;
}

static void
ivec_sort(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int *ptr = ivec_get_ptr(this);

  qsort((void *)ptr, ivec_get_size(this), sizeof(int), ivec_element_compare);
}

static void
ivec_scramble(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int *ptr = ivec_get_ptr(this);
  int size = ivec_get_size(this);
  double range = size;
  int i;

  for(i=0; i<size-1; i++)
    {
      int random = (int)(range * fts_random() / FTS_RANDOM_RANGE);
      int f = ptr[i];

      ptr[i] = ptr[i + random];
      ptr[i + random] = f;

      range -= 1.0;
    }
}

static void
ivec_rotate(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int *ptr = ivec_get_ptr(this);
  int size = ivec_get_size(this);

  if(size > 1)
    {
      int shift = 1;
      
      if(ac && fts_is_number(at))
	shift = fts_get_number_int(at);
      
      if(shift == 1)
	{
	  int f = ptr[size - 1];
	  int i;

	  for(i=size-2; i>=0; i--)
	    ptr[i + 1] = ptr[i];

	  ptr[0] = f;
	}
      else if(shift == -1)
	{
	  int f = ptr[0];
	  int i;

	  for(i=0; i<size-1; i++)
	    ptr[i] = ptr[i + 1];

	  ptr[size - 1] = f;
	}
      else
	{
	  int forward;
	  int i, j, next, end;

	  while(shift < 0)
	    shift += size;
	  
	  while(shift >= size)
	    shift -= size;
	  
	  i = 0;
	  j = shift;
	  end = shift; 

	  forward = ptr[shift];
	  ptr[shift] = ptr[0];
	  
	  while(i < end)
	    {
	      next = (j + shift) % size;
	      
	      if(next != i)
		{
		  int swap = ptr[next];

		  if(next < end)
		    end = next;

		  ptr[next] = forward;
		  forward = swap;

		  j = next;
		}
	      else
		{
		  ptr[i] = forward;

		  i++;
		  j = i;

		  forward = ptr[i];
		}
	    }
	}
    }
}

static void
ivec_size(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  if(ac > 0 && fts_is_number(at))
    {
      int size = fts_get_number_int(at);
      
      if(size >= 0)
	{
	  int old_size = this->size;
	  int i;

	  ivec_set_size(this, size);

	  /* when extending: zero new values */
	  for(i=old_size; i<size; i++)
	    this->values[i] = 0.0;	  

	  if(ivec_editor_is_open(this))
	    fts_client_send_message((fts_object_t *)this, sym_set_size, ac, at);
	}
    }
}

static void
ivec_import(fts_object_t *o, int winlet, fts_symbol_t is, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  fts_symbol_t file_name = fts_get_symbol_arg(ac, at, 0, 0);
  fts_symbol_t file_format = fts_get_symbol_arg(ac, at, 1, sym_text);

  if(!file_name)
    return;

  if(file_format == sym_text)
    {
      int size = ivec_read_atom_file(this, file_name);
      
      if(size <= 0)
	post("ivec: can't import from text file \"%s\"\n", file_name);
    }
  else
    post("ivec: unknown import file format \"%s\"\n", file_format);
}

static void
ivec_export(fts_object_t *o, int winlet, fts_symbol_t is, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  fts_symbol_t file_name = fts_get_symbol_arg(ac, at, 0, 0);
  fts_symbol_t file_format = fts_get_symbol_arg(ac, at, 1, sym_text);

  if(!file_name)
    return;

  if(file_format == sym_text)
    {
      int size = ivec_write_atom_file(this, file_name);
      
      if(size < 0)
	post("ivec: can't export to text file \"%s\"\n", file_name);
    }
  else
    post("ivec: unknown export file format \"%s\"\n", file_format);
}

static void
ivec_open_editor(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  if(!fts_object_has_id(o))
    fts_client_upload_object(o, -1);

  ivec_set_editor_open(this);
  fts_client_send_message(o, fts_s_openEditor, 0, 0);
}

static void
ivec_destroy_editor(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  ivec_set_editor_close(this);
}

static void 
ivec_close_editor(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *) o;

  if(ivec_editor_is_open(this))
    {
      ivec_set_editor_close(this);
      fts_client_send_message((fts_object_t *)this, fts_s_closeEditor, 0, 0);  
    }
}

static void
ivec_end_edit(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  fts_client_send_message(o, sym_end_edit, 0, 0);
}

static void
ivec_get_to_client(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  if(ac > 1 && fts_is_number(at))
      {
	  int first =  fts_get_number_int(at);
	  int last = fts_get_number_int(at+1);
	  ivec_append_visibles(this, first, last);
      }
  else
      ivec_send_visibles(this);
}
static void
ivec_get_pixels_to_client(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;  
  if(ac > 1 && fts_is_number(at))
      {
	  int deltax = fts_get_number_int(at);
	  int deltap = fts_get_number_int(at+1);
	  ivec_append_pixels(this, deltax, deltap);
      }
  else
      ivec_send_pixels(this);
}

static void
ivec_set_visible_window(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  
  if(ac > 1 && fts_is_number(at))
    {
      this->vsize = fts_get_number_int(at);
      this->vindex = fts_get_number_int(at+1);	  
      this->zoom = fts_get_number_float(at+2);	  
      this->pixsize = fts_get_number_int(at+3);	  
    }
}

static void
ivec_copy_by_client_request(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int start = fts_get_int(at);
  int size = fts_get_int(at + 1);
  int this_size = ivec_get_size(this);
  
  if(size > 0)
    {
      int *src, *dst;
      int i;
      
      if(!this->copy)
	this->copy = (ivec_t *)fts_object_create(ivec_type, NULL, 1, at + 1);
      else
	ivec_set_size(this->copy, size);
      
      src = ivec_get_ptr(this);
      dst = ivec_get_ptr(this->copy);
      
      if(start + size > this_size)
	size = this_size - start;
      
      for(i=0; i<size; i++)
	dst[i] = src[start + i];
    }
}

static void
ivec_cut_by_client_request(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int v_size = fts_get_int(at);
  int pix_size = fts_get_int(at + 1);
  int start = fts_get_int(at + 2);
  int copy_size;
  int *ptr;
  int i;

  ivec_copy_by_client_request(o, 0, 0, ac - 2, at + 2);
  copy_size = ivec_get_size(this->copy);

  ptr = ivec_get_ptr(this);
  
  for(i=0; i<ivec_get_size(this); i++)
    ptr[start + i] = ptr[i + start + copy_size];

  ivec_set_size(this, ivec_get_size(this) - copy_size);
  
  this->vsize = v_size;
  this->pixsize = pix_size;

  if(this->zoom < 0.5)
    ivec_send_pixels(this);

  ivec_send_visibles(this);
}

static void
ivec_paste_by_client_request(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  if(this->copy)
    {
      int start = fts_get_int(at);
      int size = fts_get_int(at + 1);
      int this_size = ivec_get_size(this);
      int copy_size = ivec_get_size(this->copy);
      int *src, *dst;
      int i;
      
      if(size == 0)
	size = copy_size;
      else if(size > copy_size)
	size = copy_size;
      
      src = ivec_get_ptr(this->copy);
      dst = ivec_get_ptr(this);
      
      if(start + size > this_size)
	size = this_size - start;
      
      for(i=0; i<size; i++)
	  dst[start+ i] = src[i];
    
      if(this->zoom<0.5) ivec_send_pixels(this);
      ivec_send_visibles(this);
    }  
}

static void
ivec_insert_by_client_request(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  if(this->copy)
    {
      int v_size = fts_get_int(at);
      int pix_size = fts_get_int(at+1);
      int start = fts_get_int(at+2);
      int copy_size = ivec_get_size(this->copy);
      int this_size;
      int *src, *dst;
      int i;

      ivec_set_size(this, ivec_get_size(this) + copy_size);
      this_size = ivec_get_size(this);

      src = ivec_get_ptr(this->copy);
      dst = ivec_get_ptr(this);

      for(i=this_size - 1; i>=start; i--)
	dst[i] = dst[i - copy_size];

      for(i=0; i<copy_size; i++)
	dst[start + i] = src[i];

      this->vsize = v_size;
      this->pixsize = pix_size;

      if(this->zoom<0.5) ivec_send_pixels(this);
      ivec_send_visibles(this);
    }
}

/********************************************************************
 *
 *  system functions
 *
 */

static void
ivec_post(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  fts_bytestream_t *stream = fts_post_get_stream(ac, at);
  int size = ivec_get_size(this);

  if(size == 0)
    fts_spost(stream, "(:ivec)");
  else if(size <= FTS_POST_MAX_ELEMENTS)
    {
      int i;
      
      fts_spost(stream, "(:ivec", size);
      
      for(i=0; i<size-1; i++)
	fts_spost(stream, "%d ", ivec_get_element(this, i));
      
      fts_spost(stream, "%d)", ivec_get_element(this, i));
    }
  else
    fts_spost(stream, "(:ivec %d)", size);
}

static void
ivec_print(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  fts_bytestream_t *stream = fts_post_get_stream(ac, at);
  int size = ivec_get_size(this);

  if(size == 0)
    fts_spost(stream, "<empty ivec>\n");
  else if(size == 1)
    fts_spost(stream, "<ivec of 1 element: %d>\n", ivec_get_element(this, 0));
  else if(size <= FTS_POST_MAX_ELEMENTS)
    {
      int i;
      
      fts_spost(stream, "<ivec of %d elements: ", size);
      
      for(i=0; i<size-1; i++)
	fts_spost(stream, "%d ", ivec_get_element(this, i));
      
      fts_spost(stream, "%d>\n", ivec_get_element(this, i));
    }
  else
    {
      int i, j;
      
      fts_spost(stream, "<ivec of %d elements>\n", size);
      fts_spost(stream, "{\n");
      
      for(i=0; i<size; i+=FTS_POST_MAX_ELEMENTS)
	{
	  int n = size - i;

	  if(n > FTS_POST_MAX_ELEMENTS)
	    n = FTS_POST_MAX_ELEMENTS;
	  
	  fts_spost(stream, "  ");
	  
	  for(j=0; j<n; j++)
	    fts_spost(stream, "%d ", ivec_get_element(this, i + j));
	  
	  fts_spost(stream, "\n");
	}
      
      fts_spost(stream, "}\n");
    }
}

static void
ivec_set_from_instance(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  ivec_t *set = (ivec_t *)fts_get_object(at);

  ivec_copy(set, this);
}

static void
ivec_dump_state(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  fts_dumper_t *dumper = (fts_dumper_t *)fts_get_object(at);      
  int size = ivec_get_size(this);
  fts_message_t *mess;
  int i;
  
  /* send size message */
  mess = fts_dumper_message_new(dumper, fts_s_size);  
  fts_message_append_int(mess, size);
  fts_dumper_message_send(dumper, mess);
  
  /* get new set message and append onset 0 */
  mess = fts_dumper_message_new(dumper, fts_s_set);
  fts_message_append_int(mess, 0);
  
  for(i=0; i<size; i++)
    {
      fts_message_append_int(mess, this->values[i]);
      
      if(fts_message_get_ac(mess) >= 256)
	{
	  fts_dumper_message_send(dumper, mess);
	  
	  /* new set message and append onset i + 1 */
	  mess = fts_dumper_message_new(dumper, fts_s_set);
	  fts_message_append_int(mess, i + 1);
	}
    }
  
  if(fts_message_get_ac(mess) > 1) 
    fts_dumper_message_send(dumper, mess);
}

static void 
ivec_dump(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  if(data_object_is_persistent(o))
    {
      fts_dumper_t *dumper = (fts_dumper_t *)fts_get_object(at);
      fts_atom_t a;

      ivec_dump_state(o, 0, 0, ac, at);

      /* save persistence flag */
      fts_set_int(&a, 1);
      fts_dumper_send(dumper, fts_s_persistence, 1, &a);      
    }

  fts_name_dump_method(o, 0, 0, ac, at);
}

static void
ivec_get_array(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  int *values = ivec_get_ptr(this);
  int size = ivec_get_size(this);
  fts_array_t *array = (fts_array_t *)fts_get_pointer(at);
  fts_atom_t *atoms;
  int i;
  
  fts_array_set_size(array, size);  
  atoms = fts_array_get_atoms(array);

  for(i=0; i<size; i++)
    fts_set_int(atoms + i, values[i]);
}

static void
ivec_set_from_array(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;

  ivec_set_size(this, ac);
  ivec_set_with_onset_from_atoms(this, 0, ac, at);
}

/*********************************************************
 *
 *  class
 *
 */
/* new/delete */

static void
ivec_init(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  
  data_object_init(o);

  this->values = 0;
  this->size = 0;
  this->alloc = 0;

  this->opened = 0; 
  this->vsize = 0; 
  this->vindex = 0;
  this->zoom = 1.0;
  this->pixsize = 1;
  this->copy = 0;

  if(ac == 0)
    ivec_set_size(this, 0);
  else if(ac == 1 && fts_is_int(at))
    {
      ivec_set_size(this, fts_get_int(at));
      ivec_set_const(this, 0);
    }
  else if(ac == 1 && fts_is_tuple(at))
    {
      fts_tuple_t *tup = (fts_tuple_t *)fts_get_object(at);
      int size = fts_tuple_get_size(tup);

      ivec_set_size(this, size);
      ivec_set_with_onset_from_atoms(this, 0, size, fts_tuple_get_atoms(tup));

      data_object_persistence_args(o);
    }
  else if(ac > 1)
    {
      ivec_set_size(this, ac);
      ivec_set_with_onset_from_atoms(this, 0, ac, at);

      data_object_persistence_args(o);
    }
  else
    fts_object_set_error(o, "bad arguments for ivec constructor");
}

static void
ivec_delete(fts_object_t *o, int winlet, fts_symbol_t s, int ac, const fts_atom_t *at)
{
  ivec_t *this = (ivec_t *)o;
  
  if(fts_object_has_id(o))
    fts_client_send_message(o, fts_s_destroyEditor, 0, 0);

  if(this->copy)
    fts_object_destroy((fts_object_t *)this->copy);

  if(this->values)
    fts_free(this->values);
}

static void
ivec_instantiate(fts_class_t *cl)
{
  fts_class_init(cl, sizeof(ivec_t), ivec_init, ivec_delete);
  
  fts_class_message_varargs(cl, fts_s_name, fts_name_set_method);
  fts_class_message_varargs(cl, fts_s_persistence, data_object_persistence);
  fts_class_message_varargs(cl, fts_s_update_gui, data_object_update_gui); 
  fts_class_message_varargs(cl, fts_s_dump_state, ivec_dump_state);
  fts_class_message_varargs(cl, fts_s_dump, ivec_dump);

  fts_class_message_varargs(cl, fts_s_post, ivec_post); 
  fts_class_message_varargs(cl, fts_s_print, ivec_print); 

  fts_class_message_varargs(cl, fts_s_set_from_instance, ivec_set_from_instance);
  fts_class_message_varargs(cl, fts_s_set_from_array, ivec_set_from_array);

  fts_class_message_varargs(cl, fts_s_get_array, ivec_get_array);

  /* graphical editor */
  fts_class_message_varargs(cl, fts_s_openEditor, ivec_open_editor);
  fts_class_message_varargs(cl, fts_s_closeEditor, ivec_close_editor);
  fts_class_message_varargs(cl, fts_s_destroyEditor, ivec_destroy_editor);
  fts_class_message_varargs(cl, fts_new_symbol("set_from_client"), ivec_set_elements);
  fts_class_message_varargs(cl, fts_new_symbol("get_from_client"), ivec_get_to_client);
  fts_class_message_varargs(cl, fts_new_symbol("get_pixels_from_client"), ivec_get_pixels_to_client);
  fts_class_message_varargs(cl, fts_new_symbol("set_visible_window"), ivec_set_visible_window);
  fts_class_message_varargs(cl, fts_new_symbol("end_edit"), ivec_end_edit);
  fts_class_message_varargs(cl, fts_new_symbol("copy_from_client"), ivec_copy_by_client_request);
  fts_class_message_varargs(cl, fts_new_symbol("paste_from_client"), ivec_paste_by_client_request);
  fts_class_message_varargs(cl, fts_new_symbol("cut_from_client"), ivec_cut_by_client_request);
  fts_class_message_varargs(cl, fts_new_symbol("insert_from_client"), ivec_insert_by_client_request);

  fts_class_message_varargs(cl, fts_new_symbol("reverse"), ivec_reverse);
  fts_class_message_varargs(cl, fts_new_symbol("rotate"), ivec_rotate);
  fts_class_message_varargs(cl, fts_new_symbol("sort"), ivec_sort);
  fts_class_message_varargs(cl, fts_new_symbol("scramble"), ivec_scramble);

  fts_class_message_varargs(cl, fts_s_fill, ivec_fill);
  fts_class_message_varargs(cl, fts_s_set, ivec_set_elements);
      
  fts_class_message_varargs(cl, fts_s_size, ivec_size);

  fts_class_message_varargs(cl, fts_s_import, ivec_import);
  fts_class_message_varargs(cl, fts_s_export, ivec_export);

  fts_class_inlet_anything(cl, 0);
}

/********************************************************************
 *
 *  config
 *
 */

void 
ivec_config(void)
{
  sym_text = fts_new_symbol("text");
  ivec_symbol = fts_new_symbol("ivec");

  sym_local = fts_new_symbol("local");

  sym_copy_client = fts_new_symbol("copy");
  sym_paste_client = fts_new_symbol("paste");
  sym_cut_client = fts_new_symbol("cut");
  sym_insert_client = fts_new_symbol("insert");

  sym_set_visibles = fts_new_symbol("setVisibles");
  sym_append_visibles = fts_new_symbol("appendVisibles");
  sym_set_pixels = fts_new_symbol("setPixels");
  sym_append_pixels = fts_new_symbol("appendPixels");
  sym_add_pixels = fts_new_symbol("addPixels");
  sym_end_edit = fts_new_symbol("endEdit");
  sym_start_edit = fts_new_symbol("startEdit");

  sym_set_size = fts_new_symbol("setSize");
  sym_set_visible_size = fts_new_symbol("setVisibleSize");

  ivec_type = fts_class_install(ivec_symbol, ivec_instantiate);
}
